from __future__ import division
import threading
import unittest
import Queue
import copy
import time

from pipevis import MainLoop


class LoopThread(threading.Thread):
    def __init__(self, loop):
        threading.Thread.__init__(self)
        self._loop = loop
        self._exception = None
    def run(self):
        try:
            self._loop()
        except BaseException, e:
            self._exception = e
    def get_exception(self):
        if self._exception is None:
            return None
        else:
            return self._exception
    
def start_loop_thread(loop):
    loop_thread = LoopThread(loop=loop)
    loop_thread.setDaemon(True) # So we can exit -- this may leave zombies
    loop_thread.start()
    return loop_thread

class TestLoop(unittest.TestCase):

    def test_process(self):
        calls = dict(init=0, final=threading.Event(), process=0)
        def init(): calls['init'] += 1
        def final():
            calls['final'].set()
        def process():
            calls['process'] += 1
            if calls['process'] >= 10:
                return MainLoop.DONE
            else:
                return MainLoop.READY
        
        loop = MainLoop(initializer=init,
                        processor=process,
                        finalizer=final,
                        poll_sleep=.01)
        loop_thread = start_loop_thread(loop=loop)

        calls['final'].wait(timeout=.5)
        self.assertTrue(calls['final'].is_set())
        self.assertEquals(calls['init'], 1)
        self.assertEquals(calls['process'], 10)

    def test_error_init(self):
        """
        Errors in init are raised by the loop.
        """
        calls = dict(final=0, process=0)
        evt = threading.Event()
        the_error = RuntimeError("AAH!")
        def init():
            try:
                raise the_error
            finally:
                evt.set()
        def final(): calls['final'] += 1
        def process(): calls['process'] += 1; return READY

        loop = MainLoop(initializer=init,
                        processor=process,
                        finalizer=final,
                        poll_sleep=.01)
        loop_thread = start_loop_thread(loop=loop)

        evt.wait(timeout=.5)
        self.assertTrue(evt.is_set())

        self.assertEquals(calls['final'], 0)
        self.assertEquals(calls['process'], 0)
        self.assertEquals(loop_thread.get_exception(), the_error)

    def test_error_process(self):
        """
        Errors in process are raised by the loop.
        """
        calls = dict(init=0, final=0)
        evt = threading.Event()
        the_error = RuntimeError("AAH!")
        def init(): calls['init'] += 1
        def final(): calls['final'] += 1
        def process():
            try:
                raise the_error
            finally:
                evt.set()

        loop = MainLoop(initializer=init,
                        processor=process,
                        finalizer=final,
                        poll_sleep=.01)
        loop_thread = start_loop_thread(loop=loop)

        evt.wait(timeout=.5)
        self.assertTrue(evt.is_set())

        # We (have to) set the event very slightly before the error
        # gets propogated up. Wait a tiny bit to abate the race conditio
        time.sleep(.001)

        self.assertEquals(calls['init'], 1)
        self.assertEquals(calls['final'], 0)
        self.assertEquals(loop_thread.get_exception(), the_error)

    def test_error_final(self):
        the_error = RuntimeError("AAH!")
        calls = dict(init=0, process=0)
        evt = threading.Event()
        def init(): calls['init'] += 1
        def final():
            try:
                raise the_error
            finally:
                evt.set()
        def process():
            calls['process'] += 1
            if calls['process'] >= 10:
                return MainLoop.DONE
            else:
                return MainLoop.READY
        
        loop = MainLoop(initializer=init,
                        processor=process,
                        finalizer=final,
                        poll_sleep=.01)
        loop_thread = start_loop_thread(loop=loop)

        evt.wait(timeout=.5)
        self.assertTrue(evt.is_set())

        self.assertEquals(calls['init'], 1)
        self.assertEquals(calls['process'], 10)
        self.assertEquals(loop_thread.get_exception(), the_error)


if __name__ == '__main__':
    unittest.main()
